#include <stdio.h>
#include <windows.h>
#include "nt_sym_link.h"
#include "privilege.h"

void print_help() {
    printf("Usage:\n"
           "  fsstat <Volume>\n"
           "Examples:\n"
           "  fsstat C:\n"
           "  fsstat \\\\Server\\Share1\n"
          );
    exit(0);
}

TranslatePath(
    PWCHAR DrvPath
    )
{
    ULONG RC;
    WCHAR szTemp[1024];

    UNICODE_STRING SrcUniName;
    UNICODE_STRING DstUniName;

    ULONG l;

    if (!GainAllPrivileges()) return 1;
    LocateNTDLLCalls();

    GetVersionEx(&OsVerInfo);
    MajorVersion = OsVerInfo.dwMajorVersion;
    MinorVersion = OsVerInfo.dwMinorVersion;

    l = swprintf(szTemp, L"\\??\\%s", DrvPath);
    if(szTemp[l] == '\\') {
        szTemp[l] = 0;
    }
    RtlInitUnicodeString(&SrcUniName, &szTemp[0]);

    printf( "  Translate path:\n");

    wprintf(L"    %s\n", DrvPath);
    RC = ParseSymLink(&SrcUniName, &DstUniName);
    if(!NT_SUCCESS(RC))
        return 1;
    ExFreePool(DstUniName.Buffer);

    return 0;
}

int
main (void)
{
    int argc;
    WCHAR** argv;
    WCHAR*  CmdLine;
    WCHAR path[MAX_PATH+1];
    WCHAR VolName[MAX_PATH+1];
    WCHAR FsName[MAX_PATH+1];
    DWORD VolSerial;
    DWORD MaxComponent;
    DWORD FsFlags;
    ULONG len;
    BOOLEAN NTFS = FALSE;
    ULONG DriveType;
    PCHAR DrvType;

    ULONG SectorsPerCluster = 0;
    ULONG BytesPerSector = 0;
    ULONG FreeClusters = 0;
    ULONG TotalClusters = 0;

    printf("Volume Information v1.0a (c) by Alter, 2004\n"
           "Home site http://www.alter.org.ua\n");

    CmdLine = GetCommandLineW();
    argv = CommandLineToArgvW(CmdLine, &argc);

    if(argc < 2) {
        print_help();
    }

    len = wcslen(argv[1]);
    if(len >= MAX_PATH-1)
        len = MAX_PATH-1;
    memcpy(&path[0], argv[1], len*sizeof(WCHAR));
    if(len && argv[1][len-1] != '\\') {
        path[len] = '\\';
        len++;
    }
    path[len] = 0;

    if(!GetVolumeInformationW(&path[0], &VolName[0], MAX_PATH, &VolSerial,
        &MaxComponent, &FsFlags, &FsName[0], MAX_PATH)) {
        wprintf(L"Volume %s is unavailable\n", path);
        return 0;
    }
    DriveType = GetDriveTypeW(path);
    GetDiskFreeSpaceW(path,
        &SectorsPerCluster,
        &BytesPerSector,
        &FreeClusters,
        &TotalClusters);

    if(!wcsncmp(FsName, L"NTFS", MAX_PATH)) {
        NTFS = TRUE;
    }

    wprintf(L"Volume %s\n"
            L"  Label:  %s\n"
            L"  Serial: %4.4x-%4.4x\n"
            L"  FS:     %s\n",
            path, VolName, VolSerial >> 16, VolSerial & 0xffff, FsName);
    switch(DriveType) {
    case DRIVE_REMOVABLE: DrvType = "Removable"   ; break;
    case DRIVE_FIXED:     DrvType = "Fixed (HDD)" ; break;
    case DRIVE_REMOTE:    DrvType = "Network"     ; break;
    case DRIVE_CDROM:     DrvType = "CD-ROM"      ; break;
    case DRIVE_RAMDISK:   DrvType = "RAM-disk"    ; break;
    default:              DrvType = "Unknown"     ; break;
    }
    printf( "  Type:   %s\n", DrvType);
    if(FsFlags & (FS_VOL_IS_COMPRESSED | FILE_READ_ONLY_VOLUME)) {
        printf( "  State: %s%s\n",
                FsFlags & FS_VOL_IS_COMPRESSED  ? " Compressed" : "",
                FsFlags & FILE_READ_ONLY_VOLUME ? " Read-only"  : ""
                );
    } else {
        printf( "  State:  Normal\n");
    }
    printf( "  Volume space:\n"
            "    Sector size:     %d bytes\n"
            "    Cluster size:    %d bytes (%d clusters)\n"
            "    Total space:     %I64d bytes (%I64dM)\n"
            "    Free space:      %I64d bytes (%I64dM)\n",
            BytesPerSector,
            SectorsPerCluster*BytesPerSector, SectorsPerCluster,
            (__int64)TotalClusters*(__int64)(SectorsPerCluster*BytesPerSector),
            (__int64)TotalClusters*(__int64)(SectorsPerCluster*BytesPerSector)/(1024*1024),
            (__int64)FreeClusters*(__int64)(SectorsPerCluster*BytesPerSector),
            (__int64)FreeClusters*(__int64)(SectorsPerCluster*BytesPerSector)/(1024*1024)
            );
    printf( "  Supported features (0x%8.8x):\n"
            "    Preserve case:     %s\n"
            "    Case sensetive:    %s\n"
            "    Unicode:           %s\n"
            "    ACLs (Security):   %s\n"
            "    File compression:  %s\n"
            "    Named streams:     %s\n"
            "    Encryption:        %s\n"
            "    Object IDs:        %s\n"
            "    Reparse points:    %s\n"
            "    Sparse files:      %s\n"
            "    Quotas:            %s\n",
            FsFlags,
            FsFlags & FS_CASE_IS_PRESERVED         ? "Yes" : "No",
            FsFlags & FS_CASE_SENSITIVE            ? "Yes" : "No",
            FsFlags & FS_UNICODE_STORED_ON_DISK    ? "Yes" : "No",
            FsFlags & FS_PERSISTENT_ACLS           ? "Yes" : "No",
            FsFlags & FS_FILE_COMPRESSION          ? "Yes" : "No",
          ((FsFlags & FILE_NAMED_STREAMS) || NTFS) ? "Yes" : "No",
            FsFlags & FILE_SUPPORTS_ENCRYPTION     ? "Yes" : "No",
            FsFlags & FILE_SUPPORTS_OBJECT_IDS     ? "Yes" : "No",
            FsFlags & FILE_SUPPORTS_REPARSE_POINTS ? "Yes" : "No",
            FsFlags & FILE_SUPPORTS_SPARSE_FILES   ? "Yes" : "No",
            FsFlags & FILE_VOLUME_QUOTAS           ? "Yes" : "No"
            );

    if(DriveType != DRIVE_REMOTE)
        TranslatePath(path);

    return 0;
}
